@@ -38,6 +38,9 @@
 #include <sys/wait.h>
 #include <assert.h>
 #include <ctype.h>
+#ifdef ANDROID
+#include <libgen.h>
+#endif
 
 #include "idlestat.h"
 #include "utils.h"
@@ -47,6 +50,7 @@
 #include "energy_model.h"
 #include "report_ops.h"
 #include "trace_ops.h"
+#include "compiler.h"
 
 #define IDLESTAT_VERSION "0.5"
 #define USEC_PER_SEC 1000000
@@ -181,8 +185,13 @@ static int display_pstates(struct report_ops *ops, void *arg, void *baseline, ch
 	return 0;
 }
 
-static int display_wakeup(struct report_ops *ops, void *arg, void *baseline, char *cpu, void *report_data)
+static int display_wakeup(struct report_ops *ops, void *arg, UNUSED void *baseline, char *cpu, void *report_data)
 {
+	/*
+	 * FIXME: This function is lacking comparison report support
+	 * When adding the feature, remember to remove the UNUSED tag
+	 * from baseline parameter.
+	 */
 	int i;
 	bool cpu_header = false;
 	struct cpuidle_cstates *cstates = arg;
@@ -271,6 +280,7 @@ int cpuidle_get_target_residency(int cpu, int state)
 	char *fpath;
 	unsigned int tr;
 	FILE *snf;
+	int ret;
 
 	if (asprintf(&fpath, CPUIDLE_STATE_TARGETRESIDENCY_PATH_FORMAT,
 				cpu, state) < 0)
@@ -283,10 +293,10 @@ int cpuidle_get_target_residency(int cpu, int state)
 		free(fpath);
 		return -1;
 	}
-	fscanf(snf, "%u", &tr);
+	ret = fscanf(snf, "%u", &tr);
 	fclose(snf);
 
-	return tr;
+	return (ret == 1) ? (int)tr : -1;
 }
 
 /**
@@ -342,7 +352,7 @@ static void release_init_pstates(struct init_pstates *initp)
 	free(initp);
 }
 
-static struct init_pstates *build_init_pstates(void)
+static struct init_pstates *build_init_pstates(struct cpu_topology *topo)
 {
 	struct init_pstates *initp;
 	int nrcpus, cpu;
@@ -368,6 +378,9 @@ static struct init_pstates *build_init_pstates(void)
 		char *fpath;
 		unsigned int *freq = &(freqs[cpu]);
 
+		if (!cpu_is_online(topo, cpu))
+			continue;
+
 		if (asprintf(&fpath, CPUFREQ_CURFREQ_PATH_FORMAT, cpu) < 0) {
 			release_init_pstates(initp);
 			return NULL;
@@ -380,7 +393,8 @@ static struct init_pstates *build_init_pstates(void)
 	return initp;
 }
 
-static void output_pstates(FILE *f, struct init_pstates *initp, int nrcpus,
+static void output_pstates(FILE *f, struct init_pstates *initp,
+				int nrcpus, struct cpu_topology *topo,
 				double ts)
 {
 	int cpu;
@@ -391,7 +405,10 @@ static void output_pstates(FILE *f, struct init_pstates *initp, int nrcpus,
 	ts_usec = (ts - ts_sec) * USEC_PER_SEC;
 
 	for (cpu = 0; cpu < nrcpus; cpu++) {
-		freq = initp? initp->freqs[cpu] : -1;
+		if (!cpu_is_online(topo, cpu))
+			continue;
+
+		freq = initp ? initp->freqs[cpu] : 0;
 		fprintf(f, "%16s-%-5d [%03d] .... %5lu.%06lu: cpu_frequency: "
 			"state=%u cpu_id=%d\n", "idlestat", getpid(), cpu,
 			ts_sec, ts_usec, freq, cpu);
@@ -406,7 +423,7 @@ static void output_pstates(FILE *f, struct init_pstates *initp, int nrcpus,
  * This function checks the array of struct cpufreq_pstate in @pstates
  * for an entry for @freq. If one if found, the index for this entry
  * is returned. If not, a new entry is inserted into the array so that
- * the frequencies are in decreasing order and the index for the new
+ * the frequencies are in increasing order and the index for the new
  * entry is returned.
  * @return: the index of the existing or newly allocated pstate struct
  */
@@ -418,7 +435,7 @@ static int alloc_pstate(struct cpufreq_pstates *pstates, unsigned int freq)
 	pstate = pstates->pstate;
 	nrfreq = pstates->max;
 
-	for (i = 0; i < nrfreq && freq <= pstate[i].freq; i++) {
+	for (i = 0; i < nrfreq && freq >= pstate[i].freq; i++) {
 		if (pstate[i].freq == freq)
 			return i;
 	}
@@ -530,10 +547,6 @@ static void open_current_pstate(struct cpufreq_pstates *ps, double time)
 static void open_next_pstate(struct cpufreq_pstates *ps, int s, double time)
 {
 	ps->current = s;
-	if (ps->idle > 0) {
-		fprintf(stderr, "Warning: opening P-state on an idle CPU\n");
-		return;
-	}
 	open_current_pstate(ps, time);
 }
 
@@ -543,11 +556,10 @@ static void close_current_pstate(struct cpufreq_pstates *ps, double time)
 	struct cpufreq_pstate *p = &(ps->pstate[c]);
 	double elapsed;
 
-	if (ps->idle > 0) {
-		fprintf(stderr, "Warning: closing P-state on an idle CPU\n");
-		return;
-	}
 	elapsed = (time - ps->time_enter) * USEC_PER_SEC;
+	if (elapsed <= 0)
+		return;
+
 	p->min_time = MIN(p->min_time, elapsed);
 	p->max_time = MAX(p->max_time, elapsed);
 	p->avg_time = AVG(p->avg_time, elapsed, p->count + 1);
@@ -555,11 +567,73 @@ static void close_current_pstate(struct cpufreq_pstates *ps, double time)
 	p->count++;
 }
 
-void cpu_change_pstate(struct cpuidle_datas *datas, int cpu,
+int record_group_freq(struct cpufreq_pstates *ps, double time,
+			      unsigned int freq)
+{
+	int cur, next;
+
+	cur = ps->current;
+	if (freq > 0)
+		next = alloc_pstate(ps, freq);
+	else
+		next = -1;
+
+	if (cur == next)
+		return 0; /* No effective change */
+
+	if (cur == -1) {
+		/* The current pstate is -1, possibly leaving idle state */
+		if (next == -1)
+			return 0; /* No known frequency to open, still idle */
+		open_next_pstate(ps, next, time);
+		return 0;
+	}
+
+	/*
+	 * The group was running, update all stats and open a new state
+	 * if needed.
+	 */
+	close_current_pstate(ps, time);
+
+	ps->current = next;
+	if (next == -1)
+		return 0;
+
+	open_current_pstate(ps, time);
+	return 0;
+}
+
+int check_pstate_composite(struct cpuidle_datas *datas, int cpu, double time)
+{
+	struct cpu_core *aff_core;
+	struct cpu_physical *aff_cluster;
+	unsigned int freq;
+
+	aff_core = cpu_to_core(cpu, datas->topo);
+	aff_cluster = cpu_to_cluster(cpu, datas->topo);
+
+	freq = core_get_highest_freq(aff_core);
+	if (aff_core->is_ht) {
+		verbose_fprintf(stderr, 5, "Core %c%d:   freq %9u, time %f\n",
+				aff_cluster->physical_id + 'A',
+				aff_core->core_id,
+				freq, time);
+	}
+	if (record_group_freq(aff_core->pstates, time, freq) == -1)
+		return -1;
+
+	freq = cluster_get_highest_freq(aff_cluster);
+	verbose_fprintf(stderr, 5, "Cluster %c: freq %9u, time %f\n",
+		aff_cluster->physical_id + 'A', freq, time);
+	return record_group_freq(aff_cluster->pstates, time, freq);
+}
+
+
+int cpu_change_pstate(struct cpuidle_datas *datas, int cpu,
 			      unsigned int freq, double time)
 {
-	struct cpufreq_pstates *ps;
-	struct cpufreq_pstate *p;
+	struct cpufreq_pstates *ps = NULL;
+	struct cpufreq_pstate *p = NULL;
 	int cur, next;
 
 	cur = get_current_pstate(datas, cpu, &ps, &p);
@@ -572,12 +646,12 @@ void cpu_change_pstate(struct cpuidle_datas *datas, int cpu,
 		 * stats unchanged
 		 */
 		ps->current = next;
-		return;
+		return 0;
 
 	case -1:
 		/* current pstate is -1, i.e. this is the first update */
 		open_next_pstate(ps, next, time);
-		return;
+		break;
 
 	case 0:
 		/* running CPU, update all stats, but skip closing current
@@ -586,13 +660,16 @@ void cpu_change_pstate(struct cpuidle_datas *datas, int cpu,
 		if (p)
 			close_current_pstate(ps, time);
 		open_next_pstate(ps, next, time);
-		return;
+		break;
 
 	default:
 		fprintf(stderr, "illegal pstate %d for cpu %d, exiting.\n",
 			cur, cpu);
 		exit(-1);
 	}
+
+	/* See if core or cluster highest frequency changed */
+	return check_pstate_composite(datas, cpu, time);
 }
 
 /**
@@ -619,21 +696,11 @@ static void merge_pstates(struct cpuidle_datas *datas,
 		percpu_a = &(datas->pstates[cpu]);
 		percpu_b = &(baseline->pstates[cpu]);
 
-		for (idx = 0; idx < percpu_a->max && idx < percpu_b->max; ) {
-			if (percpu_a->pstate[idx].freq >
-					percpu_b->pstate[idx].freq) {
-				assert(alloc_pstate(percpu_b,
-					percpu_a->pstate[idx].freq) == idx);
-				continue;
-			}
-			if (percpu_a->pstate[idx].freq <
-					percpu_b->pstate[idx].freq) {
-				assert(alloc_pstate(percpu_a,
-					percpu_b->pstate[idx].freq) == idx);
-				continue;
-			}
-			++idx;
-		}
+		for (idx = 0; idx < percpu_a->max; ++idx)
+			alloc_pstate(percpu_b, percpu_a->pstate[idx].freq);
+
+		for (idx = 0; idx < percpu_b->max; ++idx)
+			alloc_pstate(percpu_a, percpu_b->pstate[idx].freq);
 	}
 }
 
@@ -643,6 +710,9 @@ static void cpu_pstate_idle(struct cpuidle_datas *datas, int cpu, double time)
 	if (ps->current != -1)
 		close_current_pstate(ps, time);
 	ps->idle = 1;
+
+	/* See if core or cluster highest frequency changed */
+	assert(check_pstate_composite(datas, cpu, time) != -1);
 }
 
 static void cpu_pstate_running(struct cpuidle_datas *datas, int cpu,
@@ -652,6 +722,9 @@ static void cpu_pstate_running(struct cpuidle_datas *datas, int cpu,
 	ps->idle = 0;
 	if (ps->current != -1)
 		open_current_pstate(ps, time);
+
+	/* See if core or cluster highest frequency changed */
+	assert(check_pstate_composite(datas, cpu, time) != -1);
 }
 
 static int cstate_begin(double time, int state, struct cpuidle_cstates *cstates)
@@ -683,12 +756,18 @@ static void cstate_end(double time, struct cpuidle_cstates *cstates)
 	data->end = time;
 	data->duration = data->end - data->begin;
 
-	/* That happens when precision digit in the file exceed
+	/*
+	 * Duration can be < 0 when precision digit in the file exceed
 	 * 7 (eg. xxx.1000000). Ignoring the result because I don't
-	 * find a way to fix with the sscanf used in the caller
+	 * find a way to fix with the sscanf used in the caller.
+	 *
+	 * For synthetic test material, the duration may be 0.
+	 *
+	 * In both cases, do not record the entry, but do end the state
+	 * regardless.
 	 */
-	if (data->duration < 0)
-		data->duration = 0;
+	if (data->duration <= 0)
+		goto skip_entry;
 
 	/* convert to us */
 	data->duration *= USEC_PER_SEC;
@@ -716,6 +795,7 @@ static void cstate_end(double time, struct cpuidle_cstates *cstates)
 	cstate->duration += data->duration;
 	cstate->nrdata++;
 
+skip_entry:
 	/* CPU is no longer idle */
 	cstates->current_cstate = -1;
 }
@@ -762,10 +842,16 @@ int store_data(double time, int state, int cpu,
 
 	/* Update core and cluster */
 	aff_core = cpu_to_core(cpu, datas->topo);
+	state = core_get_least_cstate(aff_core);
 	if (record_cstate_event(aff_core->cstates, time, state) == -1)
 		return -1;
+
 	aff_cluster = cpu_to_cluster(cpu, datas->topo);
-	return record_cstate_event(aff_cluster->cstates, time,state);
+	state = cluster_get_least_cstate(aff_cluster);
+	if (record_cstate_event(aff_cluster->cstates, time,state) == -1)
+		return -1;
+
+	return 0;
 }
 
 static void release_datas(struct cpuidle_datas *datas)
@@ -836,13 +922,14 @@ static int store_irq(int cpu, int irqid, char *irqname,
 	return 0;
 }
 
-static void write_cstate_info(FILE *f, int cpu, char *name, int target)
+static void write_cstate_info(FILE *f, char *name, int target)
 {
 	fprintf(f, "\t%s\n", name);
 	fprintf(f, "\t%d\n", target);
 }
 
-void output_cstate_info(FILE *f, int nrcpus) {
+void output_cstate_info(FILE *f, struct cpu_topology * topo, int nrcpus)
+{
 	struct cpuidle_cstates *cstates;
 	int i, j;
 
@@ -850,9 +937,12 @@ void output_cstate_info(FILE *f, int nrcpus) {
 	assert(!is_err(cstates));
 
 	for (i=0; i < nrcpus; i++) {
+		if (!cpu_is_online(topo, i))
+			continue;
+
 		fprintf(f, "cpuid %d:\n",  i);
 		for (j=0; j < MAXCSTATE ; j++) {
-			write_cstate_info(f, i, cstates[i].cstate[j].name,
+			write_cstate_info(f, cstates[i].cstate[j].name,
 				cstates[i].cstate[j].target_residency);
 		}
 	}
@@ -863,7 +953,7 @@ void output_cstate_info(FILE *f, int nrcpus) {
 #define TRACECMD_REPORT_FORMAT "%*[^]]] %lf:%*[^=]=%u%*[^=]=%d"
 #define TRACE_FORMAT "%*[^]]] %*s %lf:%*[^=]=%u%*[^=]=%d"
 
-int get_wakeup_irq(struct cpuidle_datas *datas, char *buffer, int count)
+int get_wakeup_irq(struct cpuidle_datas *datas, char *buffer)
 {
 	int cpu, irqid;
 	char irqname[NAMELEN+1];
@@ -900,6 +990,14 @@ struct cpuidle_datas *idlestat_load(const char *filename)
 	const struct trace_ops **ops_it;
 	int ret;
 
+	/*
+	 * The linker places pointers to all entries declared with
+	 * EXPORT_TRACE_OPS into a special segment. This creates
+	 * an array of pointers preceded by trace_ops_head. Let
+	 * the static analysis tool know that we know what we are
+	 * doing.
+	 */
+	/* coverity[array_vs_singleton] */
 	for (ops_it = (&trace_ops_head)+1 ; *ops_it ; ++ops_it) {
 		assert((*ops_it)->name);
 		assert((*ops_it)->check_magic);
@@ -989,6 +1087,8 @@ int getoptions(int argc, char *argv[], struct program_options *options)
 		{ "wakeup",      no_argument,       NULL, 'w' },
 		{ "boxless-report", no_argument,    NULL, 'B' },
 		{ "csv-report",  no_argument,       NULL, 'C' },
+		{ "poll-interval", required_argument, NULL, 'I' },
+		{ "buffer-size", required_argument, NULL, 'S' },
 		{ "version",     no_argument,       NULL, 'V' },
 		{ 0, 0, 0, 0 }
 	};
@@ -1003,7 +1103,7 @@ int getoptions(int argc, char *argv[], struct program_options *options)
 
 		int optindex = 0;
 
-		c = getopt_long(argc, argv, ":b:ce:f:ho:pr:t:vwBCV",
+		c = getopt_long(argc, argv, ":b:ce:f:ho:pr:t:vwBCI:S:V",
 				long_options, &optindex);
 		if (c == -1)
 			break;
@@ -1068,6 +1168,12 @@ int getoptions(int argc, char *argv[], struct program_options *options)
 			fprintf(stderr, "-B: report type already set to %s\n",
 				options->report_type_name);
 			return -1;
+		case 'I':
+			options->tbs.poll_interval = atoi(optarg);
+			break;
+		case 'S':
+			options->tbs.percpu_buffer_size = atoi(optarg);
+			break;
 		case 0:     /* getopt_long() set a variable, just keep going */
 			break;
 		case ':':   /* missing option argument */
@@ -1123,7 +1229,7 @@ static int idlestat_file_for_each_line(const char *path, void *data,
 					int (*handler)(const char *, void *))
 {
 	FILE *f;
-	int ret;
+	int ret = 0;
 
 	if (!handler)
 		return -1;
@@ -1176,17 +1282,17 @@ static int idlestat_store(const char *path, double start_ts, double end_ts,
 	output_cpu_topo_info(cpu_topo, f);
 
 	/* output c-states information */
-	output_cstate_info(f, ret);
+	output_cstate_info(f, cpu_topo, ret);
 
 	/* emit initial pstate changes */
 	if (initp)
-		output_pstates(f, initp, initp->nrcpus, start_ts);
+		output_pstates(f, initp, initp->nrcpus, cpu_topo, start_ts);
 
 	ret = idlestat_file_for_each_line(TRACE_FILE, f, store_line);
 
 	/* emit final pstate changes */
 	if (initp)
-		output_pstates(f, NULL, initp->nrcpus, end_ts);
+		output_pstates(f, NULL, initp->nrcpus, cpu_topo, end_ts);
 
 	fclose(f);
 
@@ -1363,11 +1469,22 @@ int main(int argc, char *argv[], char *const envp[])
 		if (is_err(saved_trace_options))
 			return 1;
 
+		/*
+		 * Calculate/verify buffer size and polling trace data
+		 * interval. The interval or may may not be used to
+		 * transfer data from kernel trace buffer to some
+		 * storage media. It is needed for long eventful traces,
+		 * but is not preferred. If the user does not specify
+		 * the values, we will calculate reasonable defaults.
+		 */
+		if (calculate_buffer_parameters(options.duration, &options.tbs))
+			return 1;
+
 		/* Initialize the traces for cpu_idle and increase the
-		 * buffer size to let 'idlestat' to sleep instead of
-		 * acquiring data, hence preventing it to pertubate the
+		 * buffer size to let 'idlestat' to possibly sleep instead
+		 * of acquiring data, hence preventing it to pertubate the
 		 * measurements. */
-		if (idlestat_init_trace(options.duration))
+		if (idlestat_init_trace(options.tbs.percpu_buffer_size))
 			goto err_restore_trace_options;
 
 		/* Remove all the previous traces */
@@ -1378,7 +1495,7 @@ int main(int argc, char *argv[], char *const envp[])
 		if (get_trace_ts(&start_ts) == -1)
 			goto err_restore_trace_options;
 
-		initp = build_init_pstates();
+		initp = build_init_pstates(cpu_topo);
 
 		/* Start the recording */
 		if (idlestat_trace_enable(true))
@@ -1471,7 +1588,7 @@ int main(int argc, char *argv[], char *const envp[])
 	}
 
 	if (options.energy_model_filename)
-		calculate_energy_consumption(cpu_topo, &options);
+		calculate_energy_consumption(cpu_topo);
 
 	output_handler->close_report_file(report_data);
 
