@@ -36,7 +36,7 @@
 #include <ctype.h>
 #include <sys/stat.h>
 #include <assert.h>
-#include <values.h>
+#include <float.h>
 
 #include "list.h"
 #include "utils.h"
@@ -84,6 +84,12 @@ int add_topo_info(struct cpu_topology *topo_list, struct topology_info *info)
 	struct cpu_core     *s_core;
 	struct cpu_cpu      *s_cpu = NULL;
 	struct list_head    *ptr;
+	int *new_online_cpus;
+
+	assert(info != NULL);
+	assert(info->physical_id >= 0);
+	assert(info->core_id >= 0);
+	assert(info->cpu_id >= 0);
 
 	/* add cpu physical info */
 	ptr = check_exist_from_head(&topo_list->physical_head,
@@ -95,13 +101,13 @@ int add_topo_info(struct cpu_topology *topo_list, struct topology_info *info)
 
 		s_phy->core_num = 0;
 		s_phy->physical_id = info->physical_id;
+		s_phy->pstates = build_pstate_info(1);
 		INIT_LIST_HEAD(&s_phy->core_head);
 		INIT_LIST_HEAD(&s_phy->cpu_enum_head);
 
 		ptr = check_pos_from_head(&topo_list->physical_head,
 						s_phy->physical_id);
 		list_add_tail(&s_phy->list_physical, ptr);
-		topo_list->physical_num++;
 	} else {
 		s_phy = list_entry(ptr, struct cpu_physical, list_physical);
 	}
@@ -116,6 +122,7 @@ int add_topo_info(struct cpu_topology *topo_list, struct topology_info *info)
 		s_core->cpu_num = 0;
 		s_core->is_ht = false;
 		s_core->core_id = info->core_id;
+		s_core->pstates = build_pstate_info(1);
 		INIT_LIST_HEAD(&s_core->cpu_head);
 
 		ptr = check_pos_from_head(&s_phy->core_head, s_core->core_id);
@@ -148,9 +155,35 @@ int add_topo_info(struct cpu_topology *topo_list, struct topology_info *info)
 	/* Add to the list (really a set) of all contained cpus in s_phy */
 	list_add_tail(&s_cpu->list_phy_enum, &s_phy->cpu_enum_head);
 
+	/* Add this cpu to the list of online cpus */
+	if (topo_list->online_array_size <= info->cpu_id) {
+		new_online_cpus = realloc(topo_list->online_cpus,
+				sizeof(int) * (info->cpu_id + 1) );
+		if (new_online_cpus == NULL)
+			return error(__func__);
+
+		memset(new_online_cpus + topo_list->online_array_size,
+			0,
+			sizeof(int) *
+			(info->cpu_id + 1 - topo_list->online_array_size));
+		topo_list->online_cpus = new_online_cpus;
+		topo_list->online_array_size = info->cpu_id + 1;
+	}
+	topo_list->online_cpus[info->cpu_id] = 1;
+
 	return 0;
 }
 
+int cpu_is_online(struct cpu_topology *topo, int cpuid)
+{
+	assert(cpuid >= 0);
+
+	if (topo->online_array_size <= cpuid)
+		return 0;
+
+	return topo->online_cpus[cpuid];
+}
+
 struct cpu_physical *cpu_to_cluster(int cpuid, struct cpu_topology *topo)
 {
 	struct cpu_physical *phy;
@@ -206,6 +239,7 @@ void free_cpu_topology(struct list_head *head)
 	list_for_each_entry_safe(lphysical, n, head, list_physical) {
 		free_cpu_core_list(&lphysical->core_head);
 		list_del(&lphysical->list_physical);
+		free(lphysical->pstates->pstate);
 		free(lphysical);
 	}
 }
@@ -300,6 +334,7 @@ static struct cpu_topology *topo_folder_scan(char *path, folder_filter_t filter)
 	struct dirent dirent, *direntp;
 	struct stat s;
 	int ret;
+	int is_online;
 	struct cpu_topology *result = NULL;
 
 	dir = opendir(path);
@@ -330,9 +365,35 @@ static struct cpu_topology *topo_folder_scan(char *path, folder_filter_t filter)
 		if (filter && filter(direntp->d_name))
 			continue;
 
-		if (!strstr(direntp->d_name, "cpu"))
+		if (strncmp(direntp->d_name, "cpu", 3))
 			continue;
 
+		/*
+		 * Some cpu(s) may be offline. Before trying to figure
+		 * out the topology entry, see if the cpu is online.
+		 * The boot cpu (cpu0) is always online, so skip the check
+		 * for that.
+		 */
+		if (!strcmp(direntp->d_name, "cpu0")) {
+			is_online = 1;
+		} else {
+			ret = asprintf(&newpath, "%s/%s/%s", basedir,
+				       direntp->d_name, "online");
+			if (ret < 0)
+				goto out_free_basedir;
+
+			ret = read_int(newpath, &is_online);
+			free(newpath);
+			if (ret < 0)
+				goto out_free_basedir;
+		}
+
+		if (!is_online) {
+			verbose_fprintf(stderr, 1, "Warning: %s is offline\n",
+					direntp->d_name);
+			continue;
+		}
+
 		ret = asprintf(&newpath, "%s/%s/%s", basedir,
 				direntp->d_name, "topology");
 		if (ret < 0)
@@ -388,7 +449,6 @@ struct cpu_topology *alloc_cpu_topo_info(void)
 	if (ret == NULL)
 		return ptrerror(__func__);
 	INIT_LIST_HEAD(&ret->physical_head);
-	ret->physical_num = 0;
 
 	return ret;
 }
@@ -411,49 +471,68 @@ struct cpu_topology *read_cpu_topo_info(FILE *f, char *buf)
 		return result;
 
 	do {
+		/* Skip comment lines */
+		if (*buf == '#' || *buf == '\0') {
+			if (!fgets(buf, BUFSIZE, f))
+				goto read_error_or_eof;
+			continue;
+		}
+
+		/* Cluster line? */
 		ret = sscanf(buf, "cluster%c", &pid);
 		if (!ret)
 			break;
-
 		cpu_info.physical_id = pid - 'A';
 
-		fgets(buf, BUFSIZE, f);
+		if (!fgets(buf, BUFSIZE, f))
+			goto read_error_or_eof;
+
+		is_ht = false;
 		do {
-			ret = sscanf(buf, "\tcore%d", &cpu_info.core_id);
-			if (ret) {
-				is_ht = true;
-				fgets(buf, BUFSIZE, f);
-			} else {
-				ret = sscanf(buf, "\tcpu%d", &cpu_info.cpu_id);
-				if (ret)
-					is_ht = false;
-				else
-					break;
+			/* Skip comment lines */
+			if (*buf == '#' || *buf == '\0') {
+				if (!fgets(buf, BUFSIZE, f))
+					goto read_error_or_eof;
+				continue;
 			}
 
-			do {
-				if (!is_ht) {
-					ret = sscanf(buf, "\tcpu%d",
-						     &cpu_info.cpu_id);
-					cpu_info.core_id = cpu_info.cpu_id;
-				} else {
-					ret = sscanf(buf, "\t\tcpu%d",
-						     &cpu_info.cpu_id);
-				}
+			/* Core line? */
+			ret = sscanf(buf, "\tcore%d", &cpu_info.core_id);
+			if (ret) {
+				if (!fgets(buf, BUFSIZE, f))
+					goto read_error_or_eof;
+				is_ht = true;
+				continue;
+			}
 
-				if (!ret)
-					break;
+			/* Cpu line? */
+			if (!is_ht) {
+				ret = sscanf(buf, "\tcpu%d",
+					     &cpu_info.cpu_id);
+				cpu_info.core_id = cpu_info.cpu_id;
+			} else {
+				ret = sscanf(buf, "\t\tcpu%d",
+					     &cpu_info.cpu_id);
+			}
 
-				add_topo_info(result, &cpu_info);
+			if (!ret)
+				break;
 
-				fgets(buf, BUFSIZE, f);
-			} while (1);
+			add_topo_info(result, &cpu_info);
+
+			if (!fgets(buf, BUFSIZE, f))
+				goto read_error_or_eof;
 		} while (1);
 	} while (1);
 
 	/* output_topo_info(result); */
 
 	return result;
+
+read_error_or_eof:
+	fprintf(stderr, "Error: EOF in trace file while reading topology\n");
+	release_cpu_topo_info(result);
+	return ptrerror(NULL);
 }
 
 int release_cpu_topo_info(struct cpu_topology *topo)
@@ -463,6 +542,7 @@ int release_cpu_topo_info(struct cpu_topology *topo)
 
 	/* Free alloced memory */
 	free_cpu_topology(&topo->physical_head);
+	free(topo->online_cpus);
 	free(topo);
 
 	return 0;
@@ -503,11 +583,13 @@ void assign_baseline_in_topo(struct cpuidle_datas *datas)
 				struct cpu_physical, list_physical);
 	topo_for_each_cluster(main_phy, topo) {
 		main_phy->base_cstates = base_phy->cstates;
+		main_phy->base_pstates = base_phy->pstates;
 		/* Core loop */
 		base_core = list_first_entry(&base_phy->core_head,
 					struct cpu_core, list_core);
 		cluster_for_each_core(main_core, main_phy) {
 			main_core->base_cstates = base_core->cstates;
+			main_core->base_pstates = base_core->pstates;
 			/* Cpu loop */
 			base_cpu = list_first_entry(&base_core->cpu_head,
 					struct cpu_cpu, list_cpu);
@@ -541,16 +623,27 @@ int dump_cpu_topo_info(struct report_ops *ops, void *report_data, int (*dump)(st
 
 		sprintf(tmp, "cluster%c", s_phy->physical_id + 'A');
 
-		if (cstate)
+		if (cstate) {
 			dump(ops, s_phy->cstates, s_phy->base_cstates,
 				tmp, report_data);
+		} else {
+			dump(ops, s_phy->pstates, s_phy->base_pstates,
+				tmp, report_data);
+		}
 
 		list_for_each_entry(s_core, &s_phy->core_head, list_core) {
-			if (s_core->is_ht && cstate) {
+			if (s_core->is_ht) {
 				sprintf(tmp, "core%d", s_core->core_id);
-				dump(ops, s_core->cstates,
-					s_core->base_cstates,
-					tmp, report_data);
+
+				if (cstate) {
+					dump(ops, s_core->cstates,
+						s_core->base_cstates,
+						tmp, report_data);
+				} else {
+					dump(ops, s_core->pstates,
+						s_core->base_pstates,
+						tmp, report_data);
+				}
 			}
 
 			list_for_each_entry(s_cpu, &s_core->cpu_head,
@@ -609,21 +702,19 @@ int cluster_get_highest_freq(struct cpu_physical *clust)
 	struct cpu_cpu *cpu;
 	int cpu_pstate_index;
 	unsigned int cpu_freq;
-	unsigned int ret = ~0;
+	unsigned int ret = 0;
 
 	cluster_for_each_cpu(cpu, clust) {
 		cpu_pstate_index = cpu->pstates->current;
 		if (cpu_pstate_index < 0)
 			continue;
+		if (cpu->pstates->idle > 0)
+			continue;
 		cpu_freq = cpu->pstates->pstate[cpu_pstate_index].freq;
-		if (cpu_freq < ret)
+		if (cpu_freq > ret)
 			ret = cpu_freq;
 	}
 
-	/* It is possible we don't know anything near the start of trace */
-	if (ret == ~0)
-		ret = 0;
-
 	return ret;
 }
 
@@ -646,21 +737,17 @@ int core_get_highest_freq(struct cpu_core *core)
 	struct cpu_cpu *cpu;
 	int cpu_pstate_index;
 	unsigned int cpu_freq;
-	unsigned int ret = ~0;
+	unsigned int ret = 0;
 
 	core_for_each_cpu(cpu, core) {
 		cpu_pstate_index = cpu->pstates->current;
 		if (cpu_pstate_index < 0)
 			continue;
 		cpu_freq = cpu->pstates->pstate[cpu_pstate_index].freq;
-		if (cpu_freq < ret)
+		if (cpu_freq > ret)
 			ret = cpu_freq;
 	}
 
-	/* It is possible we don't know anything near the start of trace */
-	if (ret == ~0)
-		ret = 0;
-
 	return ret;
 }
 
@@ -693,7 +780,7 @@ static struct cpuidle_cstates *create_states(struct cpuidle_cstate *s_state)
 
 	/* Copy state information */
 	d_state = result->cstate;
-	for (i = 0; i <= MAXCSTATE; ++d_state, ++s_state, ++i) {
+	for (i = 0; i < MAXCSTATE; ++d_state, ++s_state, ++i) {
 		if (s_state->name == NULL)
 			continue;
 
@@ -791,6 +878,9 @@ int setup_topo_states(struct cpuidle_datas *datas)
 
 	/* Map cpu state arrays into topology structures */
 	for (i = 0; i < datas->nrcpus; i++) {
+		if (!cpu_is_online(topo, i))
+			continue;
+
 		s_cpu = find_cpu_point(topo, i);
 		if (s_cpu) {
 			s_cpu->cstates = &datas->cstates[i];
